@@ -25,75 +25,125 @@ use tracing::{error, info, trace, warn};
 struct Opt {
   #[structopt(long)]
   web_socket_url: String,
+
   #[structopt(
     long,
-    help = "If not specified, assumed to be the host part of <web-socket-url>",
+    help = "If not specified, assumed to be the host part of <web-socket-url>"
   )]
   xmpp_domain: Option<String>,
+
   #[structopt(long)]
   room_name: String,
+
   #[structopt(
     long,
-    help = "If not specified, assumed to be conference.<xmpp-domain>",
+    help = "If not specified, assumed to be conference.<xmpp-domain>"
   )]
   muc_domain: Option<String>,
+
   #[structopt(
     long,
-    help = "If not specified, assumed to be focus@auth.<xmpp-domain>/focus",
+    help = "If not specified, assumed to be focus@auth.<xmpp-domain>/focus"
   )]
   focus_jid: Option<String>,
+
   #[structopt(
     long,
-    default_value = "vp8",
-    help = "The video codec to negotiate support for. One of: vp8, vp9, h264",
+    default_value = "vp9",
+    help = "The video codec to negotiate support for. One of: vp9, vp8, h264"
   )]
   video_codec: String,
+
   #[structopt(long, default_value = "gst-meet")]
   nick: String,
+
   #[structopt(long)]
   region: Option<String>,
+
   #[structopt(long)]
   send_pipeline: Option<String>,
-  #[structopt(long)]
-  recv_pipeline_participant_template: Option<String>,
+
   #[structopt(
     long,
-    help = "Comma-separated endpoint IDs to select (prioritise receiving of)",
+    help = "A GStreamer pipeline which will be instantiated at startup. If an element named 'audio' is found, every remote participant's audio will be linked to it (and any 'audio' element in the recv-pipeline-participant-template will be ignored). If an element named 'video' is found, every remote participant's video will be linked to it (and any 'video' element in the recv-pipeline-participant-template will be ignored)."
+  )]
+  recv_pipeline: Option<String>,
+
+  #[structopt(
+    long,
+    help = "A GStreamer pipeline which will be instantiated for each remote participant. If an element named 'audio' is found, the participant's audio will be linked to it. If an element named 'video' is found, the participant's video will be linked to it."
+  )]
+  recv_pipeline_participant_template: Option<String>,
+
+  #[structopt(
+    long,
+    help = "Comma-separated endpoint IDs to select (prioritise receiving of)"
   )]
   select_endpoints: Option<String>,
+
   #[structopt(
     long,
-    help = "The maximum number of video streams we would like to receive",
+    help = "The maximum number of video streams we would like to receive"
   )]
   last_n: Option<u16>,
+
   #[structopt(
     long,
-    help = "The maximum height to receive video at."
+    default_value = "720",
+    help = "The maximum height we plan to send video at (used for stats only)."
   )]
-  recv_video_height: Option<u16>,
+  send_video_height: u16,
+
   #[structopt(
     long,
     help = "The video type to signal that we are sending. One of: camera, desktop"
   )]
   video_type: Option<String>,
+
+  #[structopt(
+    long,
+    default_value = "1280",
+    help = "The width to scale received video to before passing it to the recv-pipeline."
+  )]
+  recv_video_scale_width: u16,
+
+  #[structopt(
+    long,
+    default_value = "720",
+    help = "The height to scale received video to before passing it to the recv-pipeline. This will also be signalled as the maximum height that JVB should send video to us at."
+  )]
+  recv_video_scale_height: u16,
+
+  #[structopt(
+    long,
+    default_value = "200",
+    help = "The size of the jitter buffers in milliseconds. Larger values are more resilient to packet loss and jitter, smaller values give lower latency."
+  )]
+  buffer_size: u32,
+
   #[structopt(long)]
   start_bitrate: Option<u32>,
+
   #[structopt(long)]
   stereo: Option<bool>,
+
   #[structopt(short, long, parse(from_occurrences))]
   verbose: u8,
+
   #[cfg(feature = "tls-insecure")]
   #[structopt(
     long,
     help = "Disable TLS certificate verification (use with extreme caution)"
   )]
   tls_insecure: bool,
+
   #[cfg(feature = "log-rtp")]
-  #[structopt(
-    long,
-    help = "Log all RTP/RTCP packets at DEBUG level"
-  )]
+  #[structopt(long, help = "Log all RTP packets at DEBUG level (extremely verbose)")]
   log_rtp: bool,
+
+  #[cfg(feature = "log-rtp")]
+  #[structopt(long, help = "Log all RTCP packets at DEBUG level")]
+  log_rtcp: bool,
 }
 
 #[cfg(not(target_os = "macos"))]
@@ -146,7 +196,7 @@ async fn main_inner() -> Result<()> {
 
   init_gstreamer()?;
 
-  // Parse pipeline early so that we don't bother connecting to the conference if it's invalid.
+  // Parse pipelines early so that we don't bother connecting to the conference if it's invalid.
 
   let send_pipeline = opt
     .send_pipeline
@@ -154,7 +204,14 @@ async fn main_inner() -> Result<()> {
     .map(|pipeline| gstreamer::parse_bin_from_description(pipeline, false))
     .transpose()
     .context("failed to parse send pipeline")?;
-  
+
+  let recv_pipeline = opt
+    .recv_pipeline
+    .as_ref()
+    .map(|pipeline| gstreamer::parse_bin_from_description(pipeline, false))
+    .transpose()
+    .context("failed to parse recv pipeline")?;
+
   let web_socket_url: Uri = opt.web_socket_url.parse()?;
 
   let xmpp_domain = opt
@@ -198,10 +255,16 @@ async fn main_inner() -> Result<()> {
     region,
     video_codec,
     recv_pipeline_participant_template,
+    send_video_height,
+    recv_video_scale_height,
+    recv_video_scale_width,
+    buffer_size,
     start_bitrate,
     stereo,
     #[cfg(feature = "log-rtp")]
     log_rtp,
+    #[cfg(feature = "log-rtp")]
+    log_rtcp,
     ..
   } = opt;
 
@@ -214,8 +277,13 @@ async fn main_inner() -> Result<()> {
     extra_muc_features: vec![],
     start_bitrate: start_bitrate.unwrap_or(800),
     stereo: stereo.unwrap_or_default(),
+    recv_video_scale_height,
+    recv_video_scale_width,
+    buffer_size,
     #[cfg(feature = "log-rtp")]
     log_rtp,
+    #[cfg(feature = "log-rtp")]
+    log_rtcp,
   };
 
   let main_loop = glib::MainLoop::new(None, false);
@@ -224,22 +292,24 @@ async fn main_inner() -> Result<()> {
     .await
     .context("failed to join conference")?;
 
-  if opt.select_endpoints.is_some() || opt.last_n.is_some() || opt.recv_video_height.is_some() {
-    conference
-      .send_colibri_message(ColibriMessage::ReceiverVideoConstraints {
-        last_n: Some(opt.last_n.map(i32::from).unwrap_or(-1)),
-        selected_endpoints: opt
-          .select_endpoints
-          .map(|endpoints| endpoints.split(',').map(ToOwned::to_owned).collect()),
-        on_stage_endpoints: None,
-        default_constraints: opt.recv_video_height.map(|height| Constraints {
-          max_height: Some(height.into()),
-          ideal_height: None,
-        }),
-        constraints: None,
-      })
-      .await?;
-  }
+  conference
+    .set_send_resolution(send_video_height.into())
+    .await;
+
+  conference
+    .send_colibri_message(ColibriMessage::ReceiverVideoConstraints {
+      last_n: Some(opt.last_n.map(i32::from).unwrap_or(-1)),
+      selected_endpoints: opt
+        .select_endpoints
+        .map(|endpoints| endpoints.split(',').map(ToOwned::to_owned).collect()),
+      on_stage_endpoints: None,
+      default_constraints: Some(Constraints {
+        max_height: Some(opt.recv_video_scale_height.into()),
+        ideal_height: None,
+      }),
+      constraints: None,
+    })
+    .await?;
 
   if let Some(video_type) = opt.video_type {
     conference
@@ -279,6 +349,28 @@ async fn main_inner() -> Result<()> {
     conference.set_muted(MediaType::Video, true).await?;
   }
 
+  if let Some(bin) = recv_pipeline {
+    conference.add_bin(&bin).await?;
+
+    if let Some(audio_element) = bin.by_name("audio") {
+      info!(
+        "recv pipeline has an audio element, a sink pad will be requested from it for each participant"
+      );
+      conference
+        .set_remote_participant_audio_sink_element(Some(audio_element))
+        .await;
+    }
+
+    if let Some(video_element) = bin.by_name("video") {
+      info!(
+        "recv pipeline has a video element, a sink pad will be requested from it for each participant"
+      );
+      conference
+        .set_remote_participant_video_sink_element(Some(video_element))
+        .await;
+    }
+  }
+
   conference
     .on_participant(move |conference, participant| {
       let recv_pipeline_participant_template = recv_pipeline_participant_template.clone();
@@ -315,9 +407,6 @@ async fn main_inner() -> Result<()> {
             )?;
             bin.add_pad(&GhostPad::with_target(Some("audio"), &sink_pad)?)?;
           }
-          else {
-            info!("No audio sink element found in recv pipeline participant template");
-          }
 
           if let Some(video_sink_element) = bin.by_name("video") {
             let sink_pad = video_sink_element.static_pad("sink").context(
@@ -325,9 +414,6 @@ async fn main_inner() -> Result<()> {
             )?;
             bin.add_pad(&GhostPad::with_target(Some("video"), &sink_pad)?)?;
           }
-          else {
-            info!("No video sink element found in recv pipeline participant template");
-          }
 
           bin.set_property(
             "name",
@@ -335,9 +421,6 @@ async fn main_inner() -> Result<()> {
           );
           conference.add_bin(&bin).await?;
         }
-        else {
-          info!("No template for handling new participant");
-        }
 
         Ok(())
       })
